/* Copyright (c) Colorado School of Mines, 2000.*/
/* All rights reserved.                       */

/* SUAZIMUTH: $Revision: 1.8 $ ; $Date: 1999/02/23 18:18:40 $		*/

#include "su.h"
#include "segy.h"
#include "header.h"

/*********************** self documentation **********************/
char *sdoc[] = {
"									",
" SUAZIMUTH - compute trace AZIMUTH given the sx,sy,gx,gy header fields	",
"		and set a user-specified header field to this value	",
"									",
"  suazimuth <stdin >stdout [optional parameters]			",
"									",
" Required parameters:							",
" none									",
"									",
" Optional parameters:							",
" key=otrav		header field to store computed azimuths in	",
" scale=1.0		value(key) = scale * azimuth			",
" az=0			0-179.9999 deg convention, reciprocity assumed	",
"			=1 0-359.999 deg convention, points from source ",
"			to receiver					",
" sector=1.0		if set, defines output in sectors of size	",
"			sector=degrees_per_sector, the default mode is  ",
"			the full range of angles specified by az	",
" Notes:								",
" The value of header word \"key\" is computed from the values of	",
" sx,sy,gx,gy. The output field \"otrav\" was chosen arbitrarily as an	",
" example of a little-used header field, however, the user may choose 	",
" any field that is convenient for his or her application.		",

" Setting the sector=number_of_degrees_per_sector sets key field to	",
" sector number rather than an angle in degrees.			",
"  									",
" For az=0, azimuths are measured from the North, however, reciprocity  ",
" is assumed, so azimuths go from 0 to 179.9999 degrees. If sector option",
" is set, then the range is from 0 to 180/sector. 			",
"  									",
" For az=1, azimuths are measured from the North, with the assumption that",
" the direction vector points from the receiver to the source, no	",
" reciprocity is assumed, so the angles go from 0 to 359.999 degrees.	",
" If the sector option is set, then the range is from 0 to 360/sector.	",
"									",
" Type: sukeyword -o    to see the keywords and descriptions of all	",
"		       header fields.					",
"									",
" To plot midpoints, use:     su3dchart					",
"									",
NULL};

/* Credits:
 *	based on suchw, su3dchart
 *      CWP: John Stockwell and  UTulsa: Chris Liner, Oct 1998
 * 
 *  Algorithm:
 *    midpoint x  value  xm = (sx + gx)/2
 *    midpoint y  value  ym = (sy + gy)/2
 * 
 *   Azimuth will be defined as the angle, measured in degrees,
 *   turned from North, of a vector pointing to the source from the midpoint, 
 *   or from the midpoint to the source. Azimuths go from 0-179.000 degrees
 *   or from 0-180.0 degrees.
 *   
 *   value(key) = scale*[90.0 - (180.0/PI)*(atan((sy - ym)/(sx - xm))) ]
 *     or
 *   value(key) = scale*[180.0 - (180.0/PI)*(atan2((ym - sy),(xm - sx)) ]
 * 
 *   Trace header fields accessed: sx, sy, gx, gy, scalco. 
 */
/**************** end self doc ***********************************/

/* prototype for function used internally */
void computeAzimuth(cwp_String type_out, Value *val_out,
			double sx, double sy, double gx, double gy,
			double mx, double my, double scale, int az);

segy tr;

int
main(int argc, char **argv)
{
	cwp_String key;		/* output key			*/
	cwp_String type_out;	/*  type for output key		*/
	int index_out;		/*  index for output key 	*/
	Value val_out;		/* value of output key		*/

	double sx=0.0;		/* source x coordinate		*/
	double sy=0.0;		/* source y coordinate		*/
	double gx=0.0;		/* receiver x coordinate	*/
	double gy=0.0;		/* receiver y coordinate	*/
	double factor=0.0;	/* scaling factor		*/
	double scale=0.0;	/* value of scale		*/

	double mx, my;		/* x,y midpoint coordinates	*/
	int az=0;		/* azimuth convention flag	*/
	double sector=0;	/* azimuth by sectors		*/
	
	/* Initialize */
	initargs(argc, argv);
	requestdoc(1);

	/* Get parameters */
	/* get key's */
	if (!getparstring("key",&key)) 		key = "otrav";		
	if (!getpardouble("scale",&scale)) 	scale = 1.0;
	if (!getparint("az",&az))		az = 0;
	if (!getpardouble("sector",&sector)) 	sector = 1.0;

	/* absorb sector into scale */
	scale = scale/sector;

	/* get type and index values of output */
	type_out  = hdtype(key);
	index_out = getindex(key);

	/* loop over traces */
	while (gettr(&tr)) {

		/* get header values */
		sx = tr.sx;
		sy = tr.sy;
		gx = tr.gx;
		gy = tr.gy;

		/* If tr.scalco not set, use 1 as the value */
		factor = (!tr.scalco) ? 1 : tr.scalco;

		/* factor < 0 means divide; factor > 0 means to multiply */
		if (factor < 0) factor = -1/factor;

		/* compute midpoints */
		mx = (double) (0.5*(sx + gx) * factor);
		my = (double) (0.5*(sy + gy) * factor);

		/* compute the azimuths */
		computeAzimuth(type_out, &val_out, sx, sy,
				gx, gy, mx, my, scale, az);

		/* put the new header value */
		puthval(&tr, index_out, &val_out);

		/* write traces */
		puttr(&tr);

	}

	return EXIT_SUCCESS;
}

#define DRADTODEG ((double) (180.0/PI))

void computeAzimuth(cwp_String type_out, Value *val_out,
			double sx, double sy, double gx, double gy,
			double mx, double my, double scale, int az)
/************************************************************************
computeAzimuth - compute the azimuth of traces

*************************************************************************
Input:
sx		source x coordinate
sy		source y coordinate
gx		geophone x coordinate
gy		geophone y coordinate
mx		midpoint x coordinate
my		midpoint y coordinate
scale		scale factor
az		azimuth convention flag

Output:
val_out		value of output field

*************************************************************************
Notes:
Based on suchw and su3dchart. Does computations as doubles. Outputs
either as 0-179.999 or 0-359.999 degrees, or in sectors of "sector"
degrees in size.

*************************************************************************
Author: CWP: John Stockwell, 28 Oct 1998
*************************************************************************/
{
	double dval_out=0.0;	/* output value			*/

	/* compute the output */
	if (az==0) { /* 0-179.999 convention */
		/* compute angle */
 		dval_out = scale*(NINT(90.0-
				DRADTODEG*atan((sy-my)/(sx-mx)))); 

		/* Make output go from 0-179.99999 degrees. */
		if (dval_out == 180.000) dval_out = 0.0;

	} else { /* 0-359.999 */

		/* compute angle */
		dval_out = scale*DRADTODEG*atan2((sx-mx),(sy-my));

		/* Make output go from 0-359.99999 degrees. */
		if (dval_out < 0.) dval_out += 360.0;
	}

	/* Convert output to appropriate type */
	switch (*type_out) {
	case 's':
		err("can't change char header word");
	break;
	case 'h':
		val_out->h = (short) dval_out;
	break;
	case 'u':
		val_out->u = (unsigned short) dval_out;
	break;
	case 'l':
		val_out->l = (long) dval_out;
	break;
	case 'v':
		val_out->v = (unsigned long) dval_out;
	break;
	case 'i':
		val_out->i = (int) dval_out;
	break;
	case 'p':
		val_out->p = (unsigned int) dval_out;
	break;
	case 'f':
		val_out->f = (float) dval_out;
	break;
	case 'd':
		val_out->d = (double) dval_out;
	break;
	default:
		err("unknown type %s", type_out);
	break;
	}
}
