/* Copyright (c) Colorado School of Mines, 2000.*/
/* All rights reserved.                       */

/* SUEA2DF: $Revision: 1.3 $ ; $Date: 2000/09/22 16:16:40 $	    */

#include "su.h"
#include "segy.h"
#include "header.h"

/*********************** self documentation **********************/
char *sdoc[] = {
"									",
" SUEA2DF - SU version of (an)elastic anisotropic 2D finite difference 	",
"	    forward modeling						",
"									",
" Alogrithm based on Juhlin (1995, Geophys. Prosp.)			",
"       and Levander (1988, Geophysics)			                ",
" Attenuation included as in Graves (1996, BSSA)       			",
"									",
"									",
" suea2df > outfile nxl= xl= zl= rhol= vpl= vsl= [optional parameters]	",
"									",
" Required Parameters:							",
" nxl=			number of locations to define model		",
" xl=		 	model locations	in horizontal direction		",
" zl=		 	model depths					",
" rhol=		 	model rho at xl,zl				",
" vpl=		 	model vp at xl,zl				",
" vsl=		 	model vs at xl,zl				",
"									",
" Optional Parameters:							",
" dt=0.001		time sampling interval (s)			",
" ft=0.0 		first time (s)			        	",
" lt=1.0 		last time (s)				        ",
" dx=10.0      		spatial sampling intrval (m) x-coor		",
" dz=dx			spatial sampling intrval (m) z-coor		",
" xmin=-1000		first x coor (m)				",
" xmax=1000		last x coor (m)					",
" zmin=0	 	top z coor (m)  				",
" zmax=1000		bottom z coor (m)		       		",
" sx=0,sz=500		source location (m)  			        ",
" stype='p'		source type					",
" wtype='dg'		wave type					",
" ts=0.05		source length (s)				",
" favg=50		source average frequency			",
" ql=		 	model q at xl,zl				",
" qsw=0		 	switch to include attenuation =1 - include	",
" epl=		 	anisotropy ep at xl,zl				",
" dsl=		 	anisotropy ds at xl,zl				",
" anl=		 	anisotropy angle at xl,zl			",
" asw=0		 	switch to include anisotropy =1 - include	",
" msw=0			model type (=0 - field; =1 - layer)		",
" sofile=		name of source file				",
" efile=	 	name of file to output model to if desired	",
" snfile=		name of file containing for snapshots		",
" snaptime=		times of snapshots i.e. snaptime=0.1,0.2,0.3	",
" vsx=			x coordinate of vertical line of seismograms	",
" hsz=			z coordinate of horizontal line of seismograms	",
" vsfile='vsp.su'	output file for vertical line of seismograms[nz][nt]",
" hsfile='ss.su' 	output file for horizontal line of seismograms[nx][nt]",
" bc=10,10,10,10 	Top,left,bottom,right boundary condition	",
" 			=0 none						",
" 			=1 symmetry 					",
" 			=2 free surface (top only)			",
" 			>2 absorbing (value indicates width of absorbing",
"			layer	 					",
" verbose=0		=1 to print progress on screen			",
"									",
"									",
" The outfile contains information generated by the input parameters, 	",
" such as memory allocation, stability, etc. If your input file does	",
" not work, check this file first.		       			",
"									",
"									",
" The zl,vpl,vsl,rhol specify elastic constants as function of depth. 	",
" Linear interpolation and constant extrapolation is used to determine	",
" constants at depths not specified.  Values specified in depth		",
" must increase monotonically.						",
"									",
" If qsw=1 a ql array must be specified					",
"									",
" If asw=1 epl,dsl and anl arrays must be specified			",
"	epl and dsl are epsilon and delstar in Thomson (1986, Geophys.) ",
"	anl is the angle the anisotropy makes with the horizontal	",
"									",
"Output files (always generated)					",
"	hsfile								",
"	vsfile								",
"	hsfile.chd - header for hsfile					",
"	vsfile.chd - header for vsfile					",
"	hsfile.mod - model file						",
"									",
"Output files (if requested)						",
"	sofile - ascii source file					",
"	efile  - binary elastic constants file				",
"	snfile  - su format snapshots file				",
"									",
"									",
NULL};

/* 
 * Credits: UU GEOPHY Chris Juhlin 15 May 1999
 * Copyright (c) Uppsala University, 1998.
 * All rights reserved.                   
 * Parts of program use Seismic Unix Package - CSM
 * Changes - C. Juhlin
 *
 * 1. Fixed upgrading of stresses. There was an error in the coding for
 * the Tzz term, e15 was being used instead of e35. This only caused
 * probelms for dipping anisotropic layers
 *
 * 2. Added some header information for output of snapshots.
 *
 */
 
/**************** end self doc ********************************/

#define	F1	1.125
#define	F2	-0.04166667

segy tr;

/* prototypes for functions defined and used below */
int get_source(float dt, float ts, float favg, char *wtype, float *source);
static void interpovv (int nt, int ncdp, float *cdp, float **ovv, 
	float cdpt, float *ovvt);
void update_vel(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **rho, float dtx, float dtz);
void update_stress_iso(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **e11, float **e55, float dtx, float dtz);
void update_stress_ani(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **e11, float **e55, float **e33, float **e13, 
	float **e15, float **e35, float dtx, float dtz);
void add_p_source(float **txx, float** tzz, float amp, int i, int j);
void add_v_source(float **u, float **w, float amp, int i, int j);
void make_snap(float **u, float **w, float sx, float sz, 
	float dx, float dz, int nx, int nz, int k, float t,
	float xmin, float zmin,
	segy tr, FILE *sneisfp, int *wbc);
void make_seis(float **ut, float **wt, float sx, float sz, 
	float dn, float dt, int nn, int nt, float dd,
	float xmin, float zmin,
	segy tr, FILE *fp, int b1, int b2);
void get_econst(float **rho, float **e11, float **e55, float **q, int qsw, 
	float **e33, float **e13, float **e15, float **e35, int asw, 
	float dx, float dz, int nx, int nz,
	float xmin, float zmin,
	float *vlim, int msw, FILE *fp, FILE *mfp);
void abs_bc_top(float **u, float **w, float **txx, float **tzz, float **txz,
	  int nx, int nz, float *r, int *bc, int *wbc, int ifx, int ilx);
void abs_bc_left(float **u, float **w, float **txx, float **tzz, float **txz,
	  int nx, int nz, float *r, int *bc, int *wbc, int jfz, int jlz);
void abs_bc_bot(float **u, float **w, float **txx, float **tzz, float **txz,
	  int nx, int nz, float *r, int *bc, int *wbc, int ifx, int ilx);
void abs_bc_right(float **u, float **w, float **txx, float **tzz, float **txz,
	  int nx, int nz, float *r, int *bc, int *wbc, int jfz, int jlz);
void fs4s_bc_top(float **txx, float **tzz, float **txz, int nx, int nz);
void fs4v_bc_top(float **u, float **w, float **e11, float **e55, int nx, int nz);
void sym4s_bc_left(float **txx, float **tzz, float **txz, int nx, int nz);
void sym4v_bc_left(float **u, float **w, int nx, int nz);
void sym4s_bc_bot(float **txx, float **tzz, float **txz, int nx, int nz);
void sym4v_bc_bot(float **u, float **w, int nx, int nz);
void calc_area(float vmax, float dtx, float dtz, 
	int *limits, int i, int j, int k, int nx, int nz);
void rotate(float *e11, float *e33, float *e13, float *e55,
	float *e15, float *e35, float angle);
void set_layers(int nx, int nz, float *tmp, int nxl, int nzl, 
	float **pl, float **zl, float *xl, float x, float dz,
	float xmin, float zmin, int msw);
void write_chd(int nx, int nz, int nt, float dx, float dz, float dt,
	float xmin, float xmax, float zmin, float zmax,
	float sx, float sz, float favg, float ts, char *wtype, char *stype,
	int *bc, int qsw, int asw, float xz,
	FILE *fp, int comp, char *mfile);
int fwrite_chd(FILE *fp, char *s, int k);
		
/* the main program */
int
main (int argc, char **argv)
{
	float dt;		/* time sampling interval 	*/
	float ft;		/* first time sample		*/
	float lt;		/* last time sample		*/
	float dx;		/* spatial sampling interval x 	*/
	float dz;		/* spatial sampling interval z	*/
	float xmin;		/* min x coor		 	*/
	float xmax;		/* max x coor		 	*/
	float zmin;		/* min z coor		 	*/
	float zmax;		/* max z coor		 	*/
	char *stype,*wtype;	/* source, wave type		*/
	float ts;		/* source length (s)		*/
	float favg;		/* source average frequency (Hz)*/
	float sx,sz;		/* current source location (m)	*/
	int nsnap;		/* number of snapshots		*/
	float *snaptime=NULL;	/* snapshot times		*/
	int *isnap=NULL;	/* snapshot samples		*/
	char *snfile="";	/* snapshot file name 		*/
	char *sofile="";	/* source file name 		*/

	int nt;			/* number of time samples 	*/
	int nz;			/* number of z samples m+bc	*/
	int nx;			/* number of x samples m+bc	*/
	int nzm;		/* number of z samples of model */
	int nxm;		/* number of x samples of model	*/
	int nbytes;		/* number of bytes allocated	*/
	int ngrids;		/* number of bytes allocated	*/
	int i,j,k;		/* loop counters 		*/
	int msw;		/* model type */
	float **e11=NULL;	/* elastic constant c11(x,z)	*/
	float **e55=NULL;	/* elastic constant c55(x,z)	*/
	float **rho=NULL;	/* density rho(x,z)		*/
	float **q=NULL;		/* Q-factor q(x,z)		*/
	int qsw;		/* switch to include attenuation*/
	float **e33=NULL;	/* elastic constant c33(x,z)	*/
	float **e13=NULL;	/* elastic constant c13(x,z)	*/
	float **e15=NULL;	/* elastic constant c13(x,z)	*/
	float **e35=NULL;	/* elastic constant c35(x,z)	*/
	int asw;		/* switch to include anisotropy */
	float **u,**w;		/* particle velocities		*/
	float **txx,**tzz;	/* normal stresses		*/
	float **txz;		/* shear stress			*/
	float **ut,**wt;	/* time sections 		*/
	float *source;		/* source waveform		*/
	float pfac;		/* P-wave source scaling factor	*/
	int ns;			/* number of samples in source	*/
	float energy,efac;	/* energy in model		*/
	int verbose;		/* is verbose? 			*/
	float dtx,dtz;		/* fd factors			*/
	int ifx,ilx,jfz,jlz;	/* grid area for operators	*/
	int is,js;		/* source location in grid	*/
	float *vlim;		/* min and max velocity in grid	*/
	int *limits;		/* calculation area in grid	*/

	int vs2;		/* depth in samples of horiz rec line */
	int hs1;		/* horiz sample of vert rec line */
	float hsz;		/* z position of horiz receiver line */
	float vsx;		/* x position of vertical receiver line */
	char *vsfile="";	/* vert receiver seismogram line file  name */
	char *vcfile;		/* vert receiver seismogram header file name */
	char *hsfile="";	/* horiz receiver seismogram line file name */
	char *hcfile;		/* horiz receiver seismogram header file name */
	char *mfile;		/* model file name */
	char tfile[256];	/* temporary file name */

	int bc[4];		/* boundary condition flags 	*/
	int wbc[4];		/* boundary width	 	*/
	int nbc;		/* number of values given 	*/
	float *bc0;		/* taper at top boundary 	*/
	float *bc1;		/* taper at left boundary 	*/
	float *bc2;		/* taper at bottom boundary 	*/
	float *bc3;		/* taper at right boundary 	*/

	float t;		/* current time of propagation	*/

	char *efile="";		/* name of file containing constants */

	FILE *soeisfp=NULL;	/* pointer to source file  */
	FILE *sneisfp=NULL;	/* pointer to snapshot file  */
	FILE *efp=NULL;		/* constants file pointer 	*/
	FILE *hseisfp=NULL;	/* pointer to output horiz rec line file  */
	FILE *vseisfp=NULL;	/* pointer to output vert rec line file  */
	FILE *hchdfp=NULL;	/* pointer to output horiz header file  */
	FILE *vchdfp=NULL;	/* pointer to output vert header file  */
	FILE *mfp=NULL;		/* pointer to model file  */
	
	char *ri;

	ri="ri";

	
/* hook up getpar to handle the parameters */
	initargs(argc,argv);
	requestdoc(1);

/* get optional parameters */
	if (!getparfloat("dt",&dt)) dt = 0.001;
	if (!getparfloat("ft",&ft)) ft = 0.0;
	if (!getparfloat("lt",&lt)) lt = 1.0;
	if (!getparfloat("dx",&dx)) dx = 10;
	if (!getparfloat("dz",&dz)) dz = dx;
	if (!getparfloat("xmin",&xmin)) xmin = -1000;
	if (!getparfloat("xmax",&xmax)) xmax = 1000;
	if (!getparfloat("zmin",&zmin)) zmin = 0;
	if (!getparfloat("zmax",&zmax)) zmax = 1000;
	if (!getparfloat("sx",&sx)) sx = (xmin+xmax)/2;
	if (!getparfloat("sz",&sz)) sz = (zmin+zmax)/2;
	if (!getparstring("stype",&stype)) stype="p";
	if (!getparstring("wtype",&wtype)) wtype="ri";
	if (!getparfloat("ts",&ts)) ts = 0.05;
	if (!getparfloat("favg",&favg)) favg = 50;
	if (!getparint("qsw",&qsw)) qsw = 0;
	if (!getparint("asw",&asw)) asw = 0;
	if (!getparint("msw",&msw)) msw = 0;
	if (!getparint("verbose",&verbose)) verbose = 0;

	getparstring("snfile",&snfile);
	getparstring("sofile",&sofile);
		
/* determine snapshots times and sample numbers */
	nsnap = countparval("snaptime");
	if (nsnap!=0){
		snaptime = alloc1float(nsnap);
		isnap = alloc1int(nsnap);
		getparfloat("snaptime",snaptime);
		fprintf(stdout,"nsnap=%d \n",nsnap);
		for (i=0; i<nsnap; i++) isnap[i]=NINT(snaptime[i]/dt);
	}
	
	fprintf(stdout,"input\n");
	
/* if requested, open file for snapshots */
	if (*snfile!='\0') {
		if((sneisfp=fopen(snfile,"w"))==NULL)
			err("cannot open snfile=%s\n",snfile);
	}

/* if requested, open file for source */
	if (*sofile!='\0') {
		if((soeisfp=fopen(sofile,"w"))==NULL)
			err("cannot open sofile=%s\n",sofile);
	}

/* get boundary conditions */
	nbc = countparval("bc");
	if (nbc==4) {
		getparint("bc", bc);	
	} else {
		bc[0] = 10;
		bc[1] = 10;
		bc[2] = 10;
		bc[3] = 10;

		if (!((nbc==4) || (nbc==0)) ) 
			warn("Number of bc %d, using bc=10,10,10,10",bc);
	}
	for(i=0; i<4; i++){wbc[i]=2; if(bc[i]>2) wbc[i]=bc[i];}
	bc0=alloc1float(bc[0]); bc0[0]=0.95;
	bc1=alloc1float(bc[1]); bc1[0]=0.95;
	bc2=alloc1float(bc[2]); bc2[0]=0.95;
	bc3=alloc1float(bc[3]); bc3[0]=0.95;
	for(j=1; j<bc[0]; j++) bc0[j]=0.95*bc0[j-1];
	for(j=1; j<bc[1]; j++) bc1[j]=0.95*bc1[j-1];
	for(j=1; j<bc[2]; j++) bc2[j]=0.95*bc2[j-1];
	for(j=1; j<bc[3]; j++) bc3[j]=0.95*bc3[j-1];
		/*fprintf(stdout,"%d %f\n",j,bc0[j]);*/
	
	fprintf(stdout,"bc=%d %d %d %d\n",bc[0],bc[1],bc[2],bc[3]);
	fprintf(stdout,"wbc=%d %d %d %d\n",wbc[0],wbc[1],wbc[2],wbc[3]);
	
/* determine seismogram recording lines */
	if (!getparfloat("hsz",&hsz)) hsz = 0.0;
	hs1 = NINT( (hsz - zmin)/dz )+wbc[0];		/*horizontal line*/
	if (!getparfloat("vsx",&vsx)) vsx = 0.0;
	vs2 = NINT((vsx - xmin)/dx )+wbc[1];		/*vertical line*/
	if (!getparint("verbose",&verbose)) verbose = 0;
	if (!getparstring("hsfile",&hsfile)) hsfile="ss.su";
	if (!getparstring("vsfile",&vsfile)) vsfile="vsp.su";
	fprintf(stdout,"hs1=%d vs2=%d \n",hs1, vs2);
	
/* open files for seismograms */
	if((hseisfp=fopen(hsfile,"w+"))==NULL) err("cannot open hsfile=%s\n",hsfile);
	if((vseisfp=fopen(vsfile,"w+"))==NULL) err("cannot open vsfile=%s\n",vsfile);
		
/* calculate array sizes*/
	nxm=NINT((xmax-xmin)/dx)+1;	/*model size*/
	nzm=NINT((zmax-zmin)/dz)+1;
	nx=nxm+wbc[1]+wbc[3];		/*grid size including bc*/
	nz=nzm+wbc[0]+wbc[2];
	nt=NINT((lt-ft)/dt)+1;
	fprintf(stdout,"dx=%f dz=%f dt=%f\n",dx,dz,dt);
	fprintf(stdout,"nx=%d nz=%d nt=%d\n",nx,nz,nt);
	fprintf(stdout,"array sizes\n");

/* set grid parameters*/
	limits = alloc1int(4);
      	dtx=dt/dx; dtz=dt/dz;
      	energy=0;efac=1000000000000000000./((nx-4)*(nz-4));
      	pfac=1./(dx*dz);
      	is=NINT((-xmin+sx)/dx)+wbc[1]; js=NINT((-zmin+sz)/dz)+wbc[0];
 	fprintf(stdout,"is=%d js=%d\n",is,js);
 	if((sx<xmin) || (sx>xmax))
		err("sx=%f source must lie within model \n",sx);
 	if((sz<zmin) || (sz>zmax))
		err("sz=%f source must lie within model \n",sz);
    	
/* allocate space for model*/
	e11 = alloc2float(nx,nz);
	e55 = alloc2float(nx,nz);
	rho = alloc2float(nx,nz);
	ngrids=3;
	if(qsw==1) {q = alloc2float(nx,nz); ngrids+=1;}
	nbytes=ngrids*nx*nz*4;
	if(asw==1) {
		e33 = alloc2float(nx,nz);
		e13 = alloc2float(nx,nz);
		e15 = alloc2float(nx,nz);
		e35 = alloc2float(nx,nz);
		ngrids+=4;
	}
	nbytes=ngrids*nx*nz*4;
	fprintf(stdout,"%d grids allocated for model\n",ngrids);
	fprintf(stdout,"%f MBytes allocated for model\n",0.000001*nbytes);
	
/* open file for model output */
	getparstring("efile",&efile);
	if (*efile!='\0') {
		if((efp=fopen(efile,"w+"))==NULL) err("cannot open efile=%s\n",efile);
	}

/* set e_*/
	vlim = alloc1float(2);
	strcpy(tfile,hsfile);;mfile=strcat(tfile,".mod");
	if((mfp=fopen(mfile,"w+"))==NULL) err("cannot open mfile=%s\n",mfile);
	get_econst(rho,e11,e55,q,qsw,e33,e13,e15,e35,asw,
	  dx,dz,nx,nz,xmin-wbc[1]*dx,zmin-wbc[0]*dz,vlim,msw,efp,mfp);
	fclose(efp); fclose(mfp);
	fprintf(stdout,"vmin=%f vmax=%f \n",vlim[0],vlim[1]);
	fprintf(stdout,"model set\n");
	
/* calculate stability and dispersion */
	fprintf(stdout,"for stablity dt should be < %f: x-dir and %f: z-dir \n",
		0.6*dx/vlim[1],0.6*dz/vlim[1]);
	fprintf(stdout,"for low dispersion fmax should be < %f \n",vlim[0]/(5*dx));
	
		
/* get source function */
	ns=NINT(ts/dt)+1;
	source = alloc1float(ns);
	i=get_source(dt,ts,favg,wtype,source);
	fprintf(stdout,"source set\n");
	if (*sofile!='\0') {
 	  fprintf(soeisfp,"%s %s %d %d %f %f\n",wtype,stype,ns,i,favg,ts);
 	  for (i=0; i<ns; i++)fprintf(soeisfp,"%d %f\n",i,source[i]);
 	}
	pfac=1.; for (i=0; i<ns; i++) source[i]*=pfac;
	
/* allocate space for variables*/
	u = alloc2float(nx,nz);
	w = alloc2float(nx,nz);
	txx = alloc2float(nx,nz);
	tzz = alloc2float(nx,nz);
	txz = alloc2float(nx,nz);
	ngrids+=5;
	nbytes=ngrids*nx*nz*4;
	fprintf(stdout,"%f MBytes allocated for model + variables\n",0.000001*nbytes);

/* scale Q-factor */
	if(qsw==1) {
	  for (i=0; i<nx; i++){ for (j=0; j<nz; j++) {
		q[j][i]=exp(-PI*dt*favg/q[j][i]); }
	  }	
	}
	
/* zero arrays */
	for (i=0; i<nx; i++){ for (j=0; j<nz; j++) {
		u[j][i]=0;w[j][i]=0;txx[j][i]=0;tzz[j][i]=0;txz[j][i]=0; }
	}	

/* write character headers */
	strcpy(tfile,hsfile);hcfile=strcat(tfile,".chd");
	if((hchdfp=fopen(hcfile,"w+"))==NULL) err("cannot open hcfile=%s\n",hcfile);
	write_chd(nx,nz,nt,dx,dz,dt,xmin,xmax,zmin,zmax,
		sx,sz,favg,ts,wtype,stype,
		bc,qsw,asw,hsz,hchdfp,1,mfile);
	fclose(hchdfp);
	strcpy(tfile,vsfile);vcfile=strcat(tfile,".chd");
	if((vchdfp=fopen(vcfile,"w+"))==NULL) err("cannot open vcfile=%s\n",vcfile);
	write_chd(nx,nz,nt,dx,dz,dt,xmin,xmax,zmin,zmax,
		sx,sz,favg,ts,wtype,stype,
		bc,qsw,asw,vsx,vchdfp,2,mfile);
	fclose(vchdfp);

/* begin wave propagation */
	k=0;
	for (t=ft; t<=lt; t=t+dt)
{
/* calculate area for active grid */
	calc_area(vlim[1],dtx,dtz,limits,is,js,k,nx,nz);
	ifx=limits[0]; ilx=limits[1]; jfz=limits[2]; jlz=limits[3];

/* introduce velocity source */
	if (k<ns) {
	  if (strcmp(stype,"v")==0) add_v_source(u,w,source[k],is,js);
	}

/* update velocities */
	update_vel(ifx,ilx,jfz,jlz,u,w,txx,tzz,txz,rho,dtx,dtz);
	
/* apply free surface bc for velocities */
	if(bc[0]==2) fs4v_bc_top(u,w,e11,e55,nx,nz);
	
/* apply symmetry bc for velocities */
	if(bc[0]==1) sym4v_bc_left(u,w,nx,nz);
	if(bc[2]==1) sym4v_bc_bot(u,w,nx,nz);
	
/* introduce stress source */
	if (k<ns) {
	  if (strcmp(stype,"p")==0) add_p_source(txx,tzz,source[k],is,js);
	}
	
/* update stresses */
	if(asw!=1) update_stress_iso(ifx,ilx,jfz,jlz,u,w,txx,tzz,txz,e11,e55,dtx,dtz);
	if(asw==1) update_stress_ani(ifx,ilx,jfz,jlz,u,w,txx,tzz,txz,
		e11,e55,e33,e13,e15,e35,dtx,dtz);

/* apply free surface bc for stresses */
	if(bc[0]==2) fs4s_bc_top(txx,tzz,txz,nx,nz);

/* apply symmetry bc for velocities */
	if(bc[0]==1) sym4s_bc_left(txx,tzz,txz,nx,nz);
	if(bc[2]==1) sym4s_bc_bot(txx,tzz,txz,nx,nz);

/* apply absorbing bc */
	if(bc[0]>2) abs_bc_top(u,w,txx,tzz,txz,nx,nz,bc0,bc,wbc,ifx,ilx);
	if(bc[1]>2) abs_bc_left(u,w,txx,tzz,txz,nx,nz,bc1,bc,wbc,jfz,jlz);
	if(bc[2]>2) abs_bc_bot(u,w,txx,tzz,txz,nx,nz,bc2,bc,wbc,ifx,ilx);
	if(bc[3]>2) abs_bc_right(u,w,txx,tzz,txz,nx,nz,bc3,bc,wbc,jfz,jlz);
	
/* attenuation */
	if(qsw==1) {
	for (j=jfz-2; j<jlz+2; j++){ for (i=ifx-2; i<ilx+2; i++) {
		 u[j][i]*=q[j][i]; w[j][i]*=q[j][i];
		 txx[j][i]*=q[j][i]; tzz[j][i]*=q[j][i]; txz[j][i]*=q[j][i];}
	}}
	
/* "energy" */
	energy=0;
	for (i=ifx; i<ilx; i++){ for (j=jfz; j<jlz; j++) {
		energy=energy+u[j][i]*u[j][i]+w[j][i]*w[j][i]; }
	}
	
/* store wave motion on file (seismograms) */
	for (i=0 ; i < nx ; ++i) fwrite(&u[hs1][i], 4, 1, hseisfp);
	for (i=0 ; i < nx ; ++i) fwrite(&w[hs1][i], 4, 1, hseisfp);
	for (j=0 ; j < nz ; ++j) fwrite(&u[j][vs2], 4, 1, vseisfp);
	for (j=0 ; j < nz ; ++j) fwrite(&w[j][vs2], 4, 1, vseisfp);
				
/* make snapshot */
	for (i=0; i<nsnap; i++) if(k==isnap[i] && *snfile!='\0')
		make_snap(u,w,sx,sz,dx,dz,nx,nz,k,t,xmin,zmin,tr,sneisfp,wbc);

 	if((verbose==1) && (k%50==0))
		fprintf(stderr,"%d %f %f %d %d %d %d \n",
 			k,t,energy*efac,ifx,ilx,jfz,jlz);
	k=k+1;
}		
	fprintf(stdout,"propagation completed\n");
	fclose(sneisfp);

/* free space */
	free2float(u);
	free2float(w);
	free2float(txx);
	free2float(tzz);
	free2float(txz);
	fprintf(stdout,"propagation space freed\n");
	
/* read in wave motion from file and make seismograms */
	rewind(hseisfp);	/*horizontal line*/ 
	ut = alloc2float(nx,nt); wt = alloc2float(nx,nt);
	for (k=0 ; k < nt ; ++k){
	  for (i=0 ; i < nx ; ++i) fread(&ut[k][i], 4, 1, hseisfp);
	  for (i=0 ; i < nx ; ++i) fread(&wt[k][i], 4, 1, hseisfp);
	}
	make_seis(ut,wt,sx,sz,dx,dt,nx,nt,hsz,xmin,zmin,tr,hseisfp,wbc[1],wbc[3]);
	fclose(hseisfp); free2float(ut); free2float(wt);
	fprintf(stdout,"u time section output\n");
	
	rewind(vseisfp);	/*vertical line*/
	ut = alloc2float(nz,nt); wt = alloc2float(nz,nt);
	for (k=0 ; k < nt ; ++k){
	  for (j=0 ; j < nz ; ++j) fread(&ut[k][j], 4, 1, vseisfp);
	  for (j=0 ; j < nz ; ++j) fread(&wt[k][j], 4, 1, vseisfp);
	}
	make_seis(ut,wt,sx,sz,dz,dt,nz,nt,vsx,xmin,zmin,tr,vseisfp,wbc[0],wbc[2]);
	fclose(vseisfp); free2float(ut); free2float(wt);
	fprintf(stdout,"w time section output\n");
	return EXIT_SUCCESS;
}

int get_source(float dt, float ts, float favg, char *wtype, float *source)
{
	int i,ns; float t,x,xx,pi2;
	i=0; t=0; pi2=PI*PI;
	ns=NINT(ts/dt)+1;
	while (i<ns)
	{
	  if(strcmp(wtype,"ri")==0){ /*Ricker*/
	    x=favg*(t-ts/2);xx=x*x;
	    source[i]=(1-2*pi2*(xx))*exp(-pi2*xx);
	  } 							  
	  if(strcmp(wtype,"dg")==0){ /*Derivative of Gaussian*/
	    x=-4*favg*favg*pi2/log(0.1);
	    source[i]=-2*pi2*(t-ts/2)*exp(-x*(t-ts/2)*(t-ts/2));
	  } 							  
	  if(strcmp(wtype,"ga")==0){ /*Gaussian favg is 10% level for fmax*/
	    x=-favg*favg*pi2/log(0.1);
	    source[i]=exp(-x*(t-ts/2)*(t-ts/2));
	  } 							  
	  if(strcmp(wtype,"sp")==0){ /*spike*/
	    source[0]=1;
	  } 						
 	  /*fprintf(stdout,"%d %f %f\n",i,t,source[i]);*/
	  t=t+dt; i=i+1;
	}
	return(i);
}

void update_vel(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **rho, float dtx, float dtz)
{
	int i,j;
	float dtxx,dtzz,dtxz,dtzx;
      	for (j=jfz; j<=jlz; ++j) {
      	  for (i=ifx; i<=ilx; ++i) {
      	    dtxx = F1*(txx[j][i+1]-txx[j][i])
             +F2*(txx[j][i+2]-txx[j][i-1]);
      	    dtzz = F1*(tzz[j][i]-tzz[j-1][i])
      	     +F2*(tzz[j+1][i]-tzz[j-2][i]);
      	    dtxz = F1*(txz[j+1][i]-txz[j][i])
      	     +F2*(txz[j+2][i]-txz[j-1][i]);
      	    dtzx = F1*(txz[j][i]-txz[j][i-1])
      	     +F2*(txz[j][i+1]-txz[j][i-2]);
      	    u[j][i] = u[j][i] + ( dtx*dtxx + dtz*dtxz ) * rho[j][i];
      	    w[j][i] = w[j][i] + ( dtx*dtzx + dtz*dtzz ) * rho[j][i];
      	  }
      	}
}

void update_stress_iso(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **e11, float **e55, float dtx, float dtz)
{
	int i,j;
	float dux,duz,dwz,dwx;
      	for (j=jfz; j<=jlz; ++j) {
      	  for (i=ifx; i<=ilx; ++i) {
      	    dwx = F1*(w[j][i+1]-w[j][i])
             +F2*(w[j][i+2]-w[j][i-1]);
      	    duz = F1*(u[j][i]-u[j-1][i])
      	     +F2*(u[j+1][i]-u[j-2][i]);
      	    dwz = F1*(w[j+1][i]-w[j][i])
      	     +F2*(w[j+2][i]-w[j-1][i]);
      	    dux = F1*(u[j][i]-u[j][i-1])
      	     +F2*(u[j][i+1]-u[j][i-2]);
      	    txx[j][i] = txx[j][i] +
      	     e11[j][i]*dtx*dux + (e11[j][i]-2*e55[j][i])*dtz*dwz;
      	    tzz[j][i] = tzz[j][i] +
      	     e11[j][i]*dtz*dwz + (e11[j][i]-2*e55[j][i])*dtx*dux;
      	    txz[j][i] = txz[j][i] +
      	     e55[j][i]*(dtz*duz + dtx*dwx);
      	  }
      	}
}

void update_stress_ani(int ifx, int ilx, int jfz, int jlz,
	float **u, float **w, float **txx, float **tzz, float **txz,
	float **e11, float **e55, float **e33, float **e13, 
	float **e15, float **e35, float dtx, float dtz)
{
	int i,j;
	float duxm,duzm,dwzp,dwxp;
      	for (j=jfz; j<=jlz; ++j) {
      	  for (i=ifx; i<=ilx; ++i) {
      	    dwxp = F1*(w[j][i+1]-w[j][i])
             +F2*(w[j][i+2]-w[j][i-1]);
      	    duzm = F1*(u[j][i]-u[j-1][i])
      	     +F2*(u[j+1][i]-u[j-2][i]);
      	    dwzp = F1*(w[j+1][i]-w[j][i])
      	     +F2*(w[j+2][i]-w[j-1][i]);
      	    duxm = F1*(u[j][i]-u[j][i-1])
      	     +F2*(u[j][i+1]-u[j][i-2]);
      	    txx[j][i] = txx[j][i] +
      	     e11[j][i]*dtx*duxm + e13[j][i]*dtz*dwzp +
      	     e15[j][i]*(dtz*duzm + dtx*dwxp);
      	    tzz[j][i] = tzz[j][i] +
      	     e33[j][i]*dtz*dwzp + e13[j][i]*dtx*duxm +
      	     e35[j][i]*(dtz*duzm + dtx*dwxp);
      	    txz[j][i] = txz[j][i] +
      	     e15[j][i]*dtx*duxm + e35[j][i]*dtz*dwzp +
      	     e55[j][i]*(dtz*duzm + dtx*dwxp);
      	  }
      	}
}


void add_p_source(float **txx, float **tzz, float amp, int i, int j)
{
      txx[j][i] = txx[j][i] + amp;
      tzz[j][i] = tzz[j][i] + amp;
}

void add_v_source(float **u, float **w, float amp, int i, int j)
{
      w[j][i] = w[j][i] + amp*0.000001;
}

void make_snap(float **u, float **w, float sx, float sz, 
	float dx, float dz, int nx, int nz, int k, float t,
	float xmin, float zmin,
	segy tr, FILE *sneisfp, int *wbc){
	
	/* SEGY fields */
	long tracl;		/* trace number within a line */
	long tracr;		/* trace number within a reel */
	int i,j;
	
	tr.sx = sx;
	tr.selev = -sz;
	tr.trid = 1;
	tr.fldr = k;
	tr.ns = nz ;
	tr.dt = 1000 * dz;
	tr.d2 = dx ;
	tr.f1 = zmin ;
	tr.f2 = xmin ;
	tr.delrt = t ;	/*store time of snapshot*/

	if(&tracl==NULL) tracl = 0;
	tracr = 0;
	tr.duse = 2 ; /*u-component*/
	for (i=wbc[1] ; i < nx-wbc[3] ; ++i){
		++tracl; ++tracr;			
		tr.gx = xmin+dx*(i-wbc[1]);
		tr.tracl = tracl; tr.tracr = tracr;
		tr.offset = tr.gx-tr.sx ;
		for(j=wbc[0]; j<nz-wbc[2]; ++j){tr.data[j] = u[j][i];}
		fputtr(sneisfp , &tr);
	}
	tr.duse = 1 ; /*w-component*/
	for (i=wbc[1] ; i < nx-wbc[3] ; ++i){
		++tracl; ++tracr;			
		tr.gx = xmin+dx*(i-wbc[1]);
		tr.offset = tr.gx-tr.sx ;
		tr.tracl = tracl; tr.tracr = tracr;
		for(j=wbc[0]; j<nz-wbc[2]; ++j){tr.data[j] = w[j][i];}
		fputtr(sneisfp , &tr);
	}
}

void make_seis(float **ut, float **wt, float sx, float sz, 
	float dn, float dt, int nn, int nt, float dd,
	float xmin, float zmin,
	segy tr, FILE *fp, int b1, int b2){
	
	/* SEGY fields */
	long tracl=0;		/* trace number within a line */
	long tracr=0;		/* trace number within a reel */
	int i,j,k;
	
	rewind(fp); 
	tracl = tracr = 0;
	tr.sx = sx;
	tr.sdepth = 0;
	tr.ns = nt ;
	tr.d2 = dn ;
	tr.selev = -sz ;
	tr.gelev = -dd ;
	tr.f2 = xmin ;
	tr.dt = dt * 1000000.0 ;
	

	tracl = tracr = 0; k=0;
	tr.duse = 2 ; /*u-component*/
	for (i=b1 ; i < nn-b2 ; ++i){
		++tracl; ++tracr;			
		tr.tracl = tracl; tr.tracr = tracr; tr.tracf=i;
		tr.gx = xmin+dn*k ;
		tr.offset = tr.gx-tr.sx ;
		for(j=0; j<nt; ++j){tr.data[j] = ut[j][i];}
		fputtr(fp , &tr); k++;
	}
	k=0;
	tr.duse = 1 ; /*w-component*/
	for (i=b1 ; i < nn-b2 ; ++i){
		++tracl; ++tracr;			
		tr.tracl = tracl; tr.tracr = tracr; tr.tracf=i;
		tr.gx = xmin+dn*k ;
		tr.offset = tr.gx-tr.sx ;
		for(j=0; j<nt; ++j){tr.data[j] = wt[j][i];}
		fputtr(fp , &tr); k++;
	}
}

void get_econst(float **rho, float **e11, float **e55, float **q, int qsw,
	float **e33, float **e13, float **e15, float **e35, int asw, 
	float dx, float dz, int nx, int nz,
	float xmin, float zmin,
	float *vlim, int msw, FILE *fp, FILE *mfp){
		
	int nxl=0;		/* number of locations to define model	*/
	int nzl=0;		/* number of zl specified 		*/
	float *xl=NULL;		/* model locations			*/
	float **zl=NULL;	/* model depths				*/
	float **rhol=NULL;	/* model rho at xl,zl			*/
	float **vpl=NULL;	/* model vp at xl,zl sloths		*/
	float **vsl=NULL;	/* model vs at xl,zl sloths		*/
	float **ql=NULL;	/* model Q at xl,zl sloths		*/
	float **epl=NULL;	/* model (c11-c33)/2c33 at xl,zl sloths	*/
	float **dsl=NULL;	/* model del* at xl,zl sloths		*/
	float **anl=NULL;	/* model anisotropy angle at xl,zl sloths*/
	float **rhoo=NULL;	/* model rho at xl,z sloths		*/
	float **vpo=NULL;	/* model vp at xl,z			*/
	float **vso=NULL;	/* model vs at xl,z			*/
	float **qo=NULL;	/* model Q at xl,z			*/
	float **epo=NULL;	/* model (c11-c33)/2c33 at xl,z		*/
	float **dso=NULL;	/* model del* at xl,z			*/
	float **ano=NULL;	/* model anisotropy angle at xl,z	*/
	float **tmp=NULL;	/* temporary work array			*/
	float *tmp1=NULL, *tmp2=NULL;	/* temporary work arrays	*/

	int nvpl;		/* number of vpl specified 	*/
	int nvsl;		/* number of vsl specified 	*/
	int nrhol;		/* number of rhol specified 	*/
	int nql;		/* number of ql specified 	*/
	int nepl;		/* number of epl specified 	*/
	int ndsl;		/* number of dsl specified 	*/
	int nanl;		/* number of anl specified 	*/
	float zint;		/* interpolated depth 		*/
	float pint;		/* interpolated constant 	*/
	float x;		/* x surface location		*/
	int i,j;
	float b,c;		/* temporary work variables	*/
	float vmin,vmax;
	float pi;

	pi=2*acos(0.0);
	vmin=10000000;vmax=0;
	
/* check input format */
	nxl = countparval("xl");
	fprintf(stdout,"nxl=%d\n",nxl);
	if (nxl>0) {
		if (countparname("zl")!=nxl)
			err("a zl array must be specified for each xl");
		if (countparname("vpl")!=nxl)
			err("a rhol array must be specified for each xl");
		if (countparname("vpl")!=nxl)
			err("a vpl array must be specified for each xl");
		if (countparname("vsl")!=nxl)
			err("a vsl array must be specified for each xl");
		if(qsw==1){
		  if (countparname("ql")!=nxl)
			err("a ql array must be specified for each xl");}
		if(asw==1){
		  if (countparname("epl")!=nxl)
			err("an epl array must be specified for each xl");
		  if (countparname("dsl")!=nxl)
			err("a dsl array must be specified for each xl");
		  if (countparname("anl")!=nxl)
			err("an anl array must be specified for each xl");
		}
	} else {
		err("xl, zl + model parameters must be specified");
	}

	xl = ealloc1float(nxl);
	getparfloat("xl",xl);
	for (i=1; i<nxl; ++i) {
		fprintf(stdout,"%f  \n",xl[i]);
		if (xl[i]<=xl[i-1])
		err("xl values must increase monotonically");
	}

/* get model constants linearly interpolated in depth*/
	vpo = alloc2float(nz,nxl);
	vso = alloc2float(nz,nxl);
	rhoo = alloc2float(nz,nxl);
	if(qsw==1) qo = alloc2float(nz,nxl);
	if(asw==1) {
		epo = alloc2float(nz,nxl);
		dso = alloc2float(nz,nxl);
		ano = alloc2float(nz,nxl);
	}
	fprintf(stdout,"start getting model\n");
	fprintf(mfp,"msw=%d\n",msw);
	for (i=0; i<nxl; ++i) {
		nzl = countnparval(i+1,"zl");
		nvpl = countnparval(i+1,"vpl");
		nvsl = countnparval(i+1,"vsl");
		nrhol = countnparval(i+1,"rhol");
		nql = countnparval(i+1,"ql");
		nepl = countnparval(i+1,"epl");
		ndsl = countnparval(i+1,"dsl");
		nanl = countnparval(i+1,"anl");
		fprintf(stdout,"%d %d %d %d %d %d %d %d nzl nvpl nvsl nrhol nql nepl ndsl nanl\n",
		  nzl, nvpl, nvsl, nrhol, nql, nepl, ndsl, nanl);
		fprintf(mfp,"%f %d xl i\n",xl[i],i);

		if(i==0){
		 zl = ealloc2float(nzl,nxl);	
		 rhol = ealloc2float(nzl,nxl);
		 vpl = ealloc2float(nzl,nxl); 
		 vsl = ealloc2float(nzl,nxl); 
		 if(qsw==1) ql = ealloc2float(nzl,nxl); 
		 if(asw==1) epl = ealloc2float(nzl,nxl); 
		 if(asw==1) dsl = ealloc2float(nzl,nxl); 
		 if(asw==1) anl = ealloc2float(nzl,nxl); 
		 tmp1 = ealloc1float(nzl);
		 tmp2 = ealloc1float(nzl);
		}

		getnparfloat(i+1,"zl",tmp1); for (j=0;j<nzl;++j) zl[i][j]=tmp1[j];
		for (j=1; j<nzl; ++j) {
		  if (zl[i][j]<=zl[i][j-1]) err("zl must increase monotonically");}

		if (nrhol!=nzl) err("number of rhol and zl values must be equal");
		getnparfloat(i+1,"rhol",tmp2); for (j=0;j<nzl;++j) rhol[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,rhol[i][0],rhol[i][nzl-1],1,&zint,&pint); rhoo[i][j]=pint;}

		if (nvpl!=nzl) err("number of vpl and zl values must be equal");
		getnparfloat(i+1,"vpl",tmp2); for (j=0;j<nzl;++j) vpl[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,vpl[i][0],vpl[i][nzl-1],1,&zint,&pint); vpo[i][j]=pint;}

		if (nvsl!=nzl) err("number of vsl and zl values must be equal");
		getnparfloat(i+1,"vsl",tmp2); for (j=0;j<nzl;++j) vsl[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,vsl[i][0],vsl[i][nzl-1],1,&zint,&pint); vso[i][j]=pint;}

		if(qsw==1) {
		if (nql!=nzl) err("number of ql and zl values must be equal");
		getnparfloat(i+1,"ql",tmp2); for (j=0;j<nzl;++j) ql[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,ql[i][0],ql[i][nzl-1],1,&zint,&pint); qo[i][j]=pint;}
		}
		
		for (j=0; j<nzl; ++j) {
		  b=0; if(qsw==1) b=ql[i][j];
		  fprintf(mfp,"%d %f %f %f %f %f\n",j+1,zl[i][j],rhol[i][j],vpl[i][j],vsl[i][j],b);}

		if(asw==1) {
		if (nepl!=nzl) err("number of epl and zl values must be equal");
		getnparfloat(i+1,"epl",tmp2); for (j=0;j<nzl;++j) epl[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,epl[i][0],epl[i][nzl-1],1,&zint,&pint); epo[i][j]=pint;}

		if (ndsl!=nzl) err("number of dsl and zl values must be equal");
		getnparfloat(i+1,"dsl",tmp2); for (j=0;j<nzl;++j) dsl[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,dsl[i][0],dsl[i][nzl-1],1,&zint,&pint); dso[i][j]=pint;}

		if (nanl!=nzl) err("number of anl and zl values must be equal");
		getnparfloat(i+1,"anl",tmp2); for (j=0;j<nzl;++j) anl[i][j]=tmp2[j];
		for (j=0,zint=zmin; j<nz; ++j,zint+=dz) {
		  intlin(nzl,tmp1,tmp2,anl[i][0],anl[i][nzl-1],1,&zint,&pint); ano[i][j]=pint;}

		for (j=0; j<nzl; ++j){
		  fprintf(mfp,"%d %f %f %f %f \n",j+1,zl[i][j],epl[i][j],dsl[i][j],anl[i][j]);
		  anl[i][j]=anl[i][j]*pi/180.0;}
		}
		
		for(j=0; j<nzl; j++){
			if(vpl[i][j]>vmax) vmax=vpl[i][j];
			if(asw==1) {
			  if(vpl[i][j]*(1+2*epl[i][j])>vmax){
			  vmax=vpl[i][j]*(1+2*epl[i][j]);}}
			if(vsl[i][j]<vmin) vmin=vsl[i][j];}
	}
	vlim[0]=vmin;vlim[1]=vmax;

/* write model boundaries  */
	fprintf(stdout,"model boundaries\n");
	for (j=0; j<nzl; ++j) {
	  fprintf(stdout,"& && %d\n",j+1);
	  if(asw==0) for (i=0; i<nxl; ++i){
	   fprintf(stdout,"& %f %f %f %f %f \n",xl[i],zl[i][j],vpl[i][j],
		   vsl[i][j],rhol[i][j]);}
	  if(asw==1) for (j=0; j<nzl; ++j){
	   fprintf(stdout,"& %f %f %f %f %f %f %f %f\n",xl[i],zl[i][j]
	       ,vpl[i][j],vsl[i][j],rhol[i][j],epl[i][j],dsl[i][j],anl[i][j]);}
	}
	
/* interpolate model laterally (z,x) */
	tmp = alloc2float(nz,nx);
	fprintf(stdout,"start interpolating model laterally\n");
	
	vmin=10000000;vmax=0;
	for (i=0, x=xmin; i<nx; i++, x+=dx) {
	  if(msw==0) interpovv(nz,nxl,xl,rhoo,x,tmp[i]);
	  if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,rhol,zl,xl,x,dz,xmin,zmin,msw);
	  for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	  for (j=0; j<nz; j++) {rho[j][i]=tmp[i][j];
	  	if(tmp[i][j]>vmax) vmax=tmp[i][j];
	  	if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	  fprintf(stdout,"rho %f %f\n",vmin,vmax);
	vmin=10000000;vmax=0;
	for (i=0, x=xmin; i<nx; i++, x+=dx) {
	  if(msw==0) interpovv(nz,nxl,xl,vpo,x,tmp[i]);
	  if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,vpl,zl,xl,x,dz,xmin,zmin,msw);
	  for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	  for (j=0; j<nz; j++) {e11[j][i]=tmp[i][j]*tmp[i][j]*rho[j][i];
	  	if(tmp[i][j]>vmax) vmax=tmp[i][j];
	  	if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	  fprintf(stdout,"vp %f %f\n",vmin,vmax);
	vmin=10000000;vmax=0;
	for (i=0, x=xmin; i<nx; i++, x+=dx) {
	  if(msw==0) interpovv(nz,nxl,xl,vso,x,tmp[i]);
	  if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,vsl,zl,xl,x,dz,xmin,zmin,msw);
	  for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	  for (j=0; j<nz; j++) {e55[j][i]=tmp[i][j]*tmp[i][j]*rho[j][i];
	  	if(tmp[i][j]>vmax) vmax=tmp[i][j];
	  	if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	  fprintf(stdout,"vs %f %f\n",vmin,vmax);
	for (i=0, x=xmin; i<nx; i++, x+=dx) {
	  for (j=0; j<nz; j++) rho[j][i]=1/rho[j][i];}
	  
	if(qsw==1){
	  vmin=10000000;vmax=0;
	  for (i=0, x=xmin; i<nx; i++, x+=dx) {
	    if(msw==0) interpovv(nz,nxl,xl,qo,x,tmp[i]);
	    if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,ql,zl,xl,x,dz,xmin,zmin,msw);
	    for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	    for (j=0; j<nz; j++) {q[j][i]=tmp[i][j];
	  	if(tmp[i][j]>vmax) vmax=tmp[i][j];
	  	if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	    fprintf(stdout,"q %f %f\n",vmin,vmax);
	}
	    
	if(asw==1){
	  vmin=10000000;vmax=0;
	  for (i=0, x=xmin; i<nx; i++, x+=dx)
	  {
	    if(msw==0) interpovv(nz,nxl,xl,epo,x,tmp[i]);
	    if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,epl,zl,xl,x,dz,xmin,zmin,msw);
	    for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	    for (j=0; j<nz; j++) e33[j][i]=e11[j][i];
	    for (j=0; j<nz; j++) {e11[j][i]=e33[j][i]*(1+2*tmp[i][j]);
	  	if(tmp[i][j]>vmax) vmax=tmp[i][j];
	  	if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	    fprintf(stdout,"ep %f %f\n",vmin,vmax);
	  vmin=10000000;vmax=0;
	  for (i=0, x=xmin; i<nx; i++, x+=dx)
	  {
	    if(msw==0) interpovv(nz,nxl,xl,dso,x,tmp[i]);
	    if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,dsl,zl,xl,x,dz,xmin,zmin,msw);
	    for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	    for (j=0; j<nz; j++)
	    {
	      b=2*e55[j][i];
	      c=e55[j][i]*e55[j][i]-tmp[i][j]*e33[j][i]*e33[j][i]-
	        0.5*(e33[j][i]-e55[j][i])*(e11[j][i]+e33[j][i]-2*e55[j][i]);
	      e13[j][i]=0.5*(-b+sqrt(b*b-4*c));
	      if(tmp[i][j]>vmax) vmax=tmp[i][j];
	      if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	    fprintf(stdout,"ds %f %f\n",vmin,vmax);
	  vmin=10000000;vmax=0;
	  for (i=0, x=xmin; i<nx; i++, x+=dx)
	  {
	    if(msw==0) interpovv(nz,nxl,xl,ano,x,tmp[i]);
	    if(msw!=0) set_layers(nx,nz,tmp[i],nxl,nzl,anl,zl,xl,x,dz,xmin,zmin,msw);
	    for (j=0; j<nz; ++j) fwrite(&tmp[i][j], 4, 1, fp);
	    for (j=0; j<nz; j++)
	    {
	      e15[j][i]=0.0;
	      e35[j][i]=0.0;
	      rotate(&e11[j][i],&e33[j][i],&e13[j][i],&e55[j][i],&e15[j][i]
	        ,&e35[j][i],tmp[i][j]);
	      if(tmp[i][j]>vmax) vmax=tmp[i][j];
	      if(tmp[i][j]<vmin) vmin=tmp[i][j];}}
	    fprintf(stdout,"an %f %f\n",vmin,vmax);
	    
	}

	free2float(zl);
	free2float(vpl);
	free2float(vsl);
	free2float(rhol);
	if(qsw==1) free2float(ql);
	if(asw==1) free2float(epl);
	if(asw==1) free2float(dsl);
	if(asw==1) free2float(anl);
	
	free2float(vpo);
	free2float(vso);
	free2float(rhoo);
	if(qsw==1) free2float(qo);
	if(asw==1) free2float(epo);
	if(asw==1) free2float(dso);
	if(asw==1) free2float(ano);
	free2float(tmp);
	free1float(tmp1);
	free1float(tmp2);
	}

/* linearly interpolate/extrapolate sloth between model points (x,t) */
	static void interpovv (int nt, int ncdp, float *cdp, float **ovv, 
	float cdpt, float *ovvt){

	static int index=0;
	int it;
	float a1,a2;

	/* if before first cdp, constant extrapolate */
	if (cdpt<=cdp[0]) {
		for (it=0; it<nt; ++it)
			ovvt[it] = ovv[0][it];
	
	/* else if beyond last cdp, constant extrapolate */
	} else if (cdpt>=cdp[ncdp-1]) {
		for (it=0; it<nt; ++it)
			ovvt[it] = ovv[ncdp-1][it];
	
	/* else, linearly interpolate */
	} else {
		xindex(ncdp,cdp,cdpt,&index);
		a1 = (cdp[index+1]-cdpt)/(cdp[index+1]-cdp[index]);
		a2 = (cdpt-cdp[index])/(cdp[index+1]-cdp[index]);
		for (it=0; it<nt; ++it)
			ovvt[it] = a1*ovv[index][it]+a2*ovv[index+1][it];
	}
	}

/* absorbing bc for top*/	
void abs_bc_top(float **u, float **w, float **txx, float **tzz, float **txz,
	  int nx, int nz, float *r, int *bc, int *wbc, int ifx, int ilx)
	{
	int i,j; float fac;
		
	for (j=0; j<wbc[0]; ++j){
	  fac=r[(wbc[0]-j-1)];
	  for (i=ifx; i<ilx; ++i){
	    u[j][i]*=fac; w[j][i]*=fac;
	    txx[j][i]*=fac; tzz[j][i]*=fac; txz[j][i]*=fac;}}
	}

/* absorbing bc for left side*/	
void abs_bc_left(float **u, float **w, float **txx, float **tzz, float **txz,
	int nx, int nz, float *r, int *bc, int *wbc, int jfz, int jlz)
	{
	int i,j; float fac;
		
	for (j=jfz; j<jlz; ++j){
	  for (i=0; i<wbc[1]; ++i){
	    fac=r[(wbc[1]-i-1)];
	    u[j][i]*=fac; w[j][i]*=fac;
	    txx[j][i]*=fac; tzz[j][i]*=fac; txz[j][i]*=fac;}}
	}

/* absorbing bc for bottom*/	
	void abs_bc_bot(float **u, float **w, float **txx, float **tzz, float **txz,
	int nx, int nz, float *r, int *bc, int *wbc, int ifx, int ilx)
	{
	int i,j; float fac;
		
	for (j=nz-wbc[2]; j<nz; ++j){
	  fac=r[(j-nz+wbc[2])];
	  for (i=ifx; i<ilx; ++i){
	    u[j][i]*=fac; w[j][i]*=fac;
	    txx[j][i]*=fac; tzz[j][i]*=fac; txz[j][i]*=fac;}}	    
	}

/* absorbing bc for right side*/	
void abs_bc_right(float **u, float **w, float **txx, float **tzz, float **txz,
	int nx, int nz, float *r, int *bc, int *wbc, int jfz, int jlz)
	{
	int i,j; float fac;
		
	for (j=jfz; j<jlz; ++j){
	  for (i=nx-wbc[3]; i<nx; ++i){
	    fac=r[(i-nx+wbc[3])];
	    u[j][i]*=fac; w[j][i]*=fac;
	    txx[j][i]*=fac; tzz[j][i]*=fac; txz[j][i]*=fac;}}
	}
	
/* Free surface boundary condition where stresses are set to zero
   for the order 4 FD operater on a staggered grid */
void fs4s_bc_top(float **txx, float **tzz, float **txz, int nx, int nz)
	{
	int i;
		
	for (i=2; i<nx-2; ++i){
      		txx[1][i]=0.; txx[0][i]=0.; 
      		txz[1][i]=0.; txz[0][i]=0.; txz[2][i]=0.;
      		tzz[1][i]=-tzz[2][i]; tzz[0][i]=-tzz[3][i];}
	}

/* Free surface boundary condition where velocities satisfy stress free
   surface for the order 4 FD operater on a staggered grid

  ** requires anisotropy symmetry parallel to free surface in 
  ** current implementation */

void fs4v_bc_top(float **u, float **w, float **e11, float **e55, int nx, int nz)
	{
	int i;
		
	for (i=2; i<nx-2; ++i){
	  u[0][i]=u[2][i]+
	    2*( F1*(w[2][i+1]-w[2][i]) + F2*(w[2][i+2]-w[2][i-1]) );
	  w[0][i]=w[2][i]+
/*	    2*((e11[2][i]-2*e55[2][i])/e11[2][i])*	*/
	    2*(1-2*(e55[2][i]/e11[2][i]))*	
	    ( F1*(u[0][i]-u[0][i-1]) + F2*(u[0][i+1]-u[0][i-2]) );
	  u[1][i]=u[0][i]+
	    2*( F1*(w[2][i+1]-w[2][i]) + F2*(w[2][i+2]-w[2][i-1]) );
	  w[1][i]=w[0][i]+
/*	    2*((e11[2][i]-2*e55[2][i])/e11[2][i])*	*/
	    2*(1-2*(e55[2][i]/e11[2][i]))*	
	    ( F1*(u[1][i]-u[1][i-1]) + F2*(u[1][i+1]-u[1][i-2]) );
	    }
	}

/* Stress symmetry boundary condition for the LHS boundary
   for the order 4 FD operater on a staggered grid */
void sym4s_bc_left(float **txx, float **tzz, float **txz, int nx, int nz)
	{
	int j;
		
	for (j=0; j<nz; ++j){
      		txx[j][1]=-txx[j][3]; txx[j][0]=-txx[j][4]; 
      		tzz[j][1]=tzz[j][3];  tzz[j][0]=tzz[j][4];
      		txz[j][1]=-txz[j][2]; txz[j][0]=-txz[j][3];}
	}

/* Velocity symmetry boundary condition for the LHS boundary
   for the order 4 FD operater on a staggered grid */
void sym4v_bc_left(float **u, float **w, int nx, int nz)
	{
	int j;
		
	for (j=0; j<nz; ++j){
      		u[j][1]=-u[j][2]; u[j][0]=-u[j][2]; 
      		w[j][1]=w[j][3]; w[j][0]=w[j][4];}
	}

/* Stress symmetry boundary condition for the BOTTOM boundary
   for the order 4 FD operater on a staggered grid */
void sym4s_bc_bot(float **txx, float **tzz, float **txz, int nx, int nz)
	{
	int i;
		
	for (i=2; i<nx-2; ++i){
      		txx[nz-2][i]=txx[nz-4][i];  txx[nz-1][i]=txx[nz-5][i]; 
      		tzz[nz-2][i]=-tzz[nz-4][i]; tzz[nz-1][i]=-tzz[nz-5][i];
      		txz[nz-2][i]=-txz[nz-3][i]; txz[nz-1][i]=-txz[nz-4][i];}
	}

/* Velocity symmetry boundary condition for the BOTTOM boundary
   for the order 4 FD operater on a staggered grid */
void sym4v_bc_bot(float **u, float **w, int nx, int nz)
	{
	int i;
		
	for (i=2; i<nx-2; ++i){
      		u[nz-2][i]=u[nz-4][i]; u[nz-1][i]=u[nz-5][i]; 
      		w[nz-2][i]=-w[nz-3][i];w[nz-1][i]=-w[nz-4][i];}
	}

void calc_area(float vmax, float dtx, float dtz, 
	int *limits, int i, int j, int k, int nx, int nz)
{
      limits[0] = i - NINT(vmax*k*dtx) - 10;
      limits[1] = i + NINT(vmax*k*dtx) + 10;
      limits[2] = j - NINT(vmax*k*dtz) - 10;
      limits[3] = j + NINT(vmax*k*dtz) + 10;
      if (limits[0] < 2) limits[0] = 2;
      if (limits[1] > nx-3) limits[1] = nx-3;
      if (limits[2] < 2) limits[2] = 2;
      if (limits[3] > nz-3) limits[3] = nz-3;      
}


void rotate(float *e11, float *e33, float *e13, float *e55,
	float *e15, float *e35, float angle)
{
	float e11p,e33p,e13p,e55p,e15p,e35p;
	float c,s;
	
	if(angle!=0.0){
		
	c=cos(angle); s=sin(angle);
	
	e11p= c*c*c*c*(*e11) + c*c*s*s*(2*(*e13)+4*(*e55)) + s*s*s*s*(*e33);
	e33p= c*c*c*c*(*e33) + c*c*s*s*(2*(*e13)+4*(*e55)) + s*s*s*s*(*e11);
	e13p= c*c*c*c*(*e13) + c*c*s*s*(*e11+*e33-4*(*e55)) + s*s*s*s*(*e13);
	e55p= c*c*c*c*(*e55) + c*c*s*s*(*e11+*e33-2*(*e55)-2*(*e13)) + s*s*s*s*(*e55);
	e15p= c*c*c*s*(-*e11+*e13+2*(*e55)) + c*s*s*s*(*e33-*e13-2*(*e55));
	e35p= c*s*s*s*(-*e11+*e13+2*(*e55)) + c*c*c*s*(*e33-*e13-2*(*e55));
	 
	*e11=e11p;
	*e33=e33p;
	*e13=e13p;
	*e55=e55p;
	*e15=e15p;
	*e35=e35p;
	}
}

void set_layers(int i, int nz, float *tmp, int nxl, int nzl, 
	float **pl, float **zl, float *xl, float x, float dz,
	float xmin, float zmin, int msw)
{
	int j,i1,j1,j2,jl;
	float z=0.0,x1=0.0,x2=0.0,p=0.0;
	static int il;
	
/*	if(NINT(x)%100==0)
	    fprintf(stdout,"msw %d %d %d %f \n",msw,nzl,nxl,x);*/

	j1=0; i1=0;
	
	  for (jl=0; jl<=nzl; ++jl){
	  
	    /* if before first location, constant extrapolate */
	    if (x<=xl[0]) {z=zl[0][jl];
	    
	    /* else if beyond last location constant extrapolate */
	    } else if (x>=xl[nxl-1]) {z=zl[nxl-1][jl];
	    
	    /* else, linearly interpolate */
	    /*if ( x>xl[0] & x<xl[nxl-1] ) {*/
	    } else {
	      xindex(nxl,xl,x,&il);
	      x1 = (xl[il+1]-x)/(xl[il+1]-xl[il]);
	      x2 = (x-xl[il])/(xl[il+1]-xl[il]);
	      if(jl<nzl) z = x1*zl[il][jl]+x2*zl[il+1][jl];
	    }  

	    /* choose constant layer velocity */
	    if(msw==1) p = pl[il][jl];
	    /* interpolate laterally layer velocity */
	    if(msw==2) p = x1*pl[il][jl]+x2*pl[il+1][jl];
	    
	    j2=NINT((z-zmin)/dz); if(jl==nzl) j2=nz;
	    if(j2>nz) j2=nz;
	    if(j1<0) j2=0;
/*	if(NINT(x)%100==0)
	    fprintf(stdout,"%d %d %d %d %f %f\n",il,jl,j1,j2,z,p);*/
	    
	    for (j=j1; j<j2; ++j) tmp[j]=p;
	    j1=j2;
	  }
}


void write_chd(int nx, int nz, int nt, float dx, float dz, float dt,
	float xmin, float xmax, float zmin, float zmax,
	float sx, float sz, float favg, float ts, char *wtype, char *stype,
	int *bc, int qsw, int asw, float xz,
	FILE *fp, int comp, char *mfile)
{
	int k=1;
	char s[80];
	k=fwrite_chd(fp,"SUEA2DF",k); 
	k=fwrite_chd(fp,"Department of Earth Sciences",k);
	k=fwrite_chd(fp,"Uppsala University",k);
	k=fwrite_chd(fp,"Date:",k);
	k=fwrite_chd(fp," ",k);
	if(comp==1) {sprintf(s,"Horizontal line recording at z=%f",xz);
	  k=fwrite_chd(fp,s,k);}
	if(comp==2) {sprintf(s,"Vertical line recording at x=%f",xz);
	  k=fwrite_chd(fp,s,k);}
	sprintf(s,"nx=%d nz=%d nt=%d dx=%f dz=%f dt=%f",nx,nz,nt,dx,dz,dt);
	  k=fwrite_chd(fp,s,k);
	sprintf(s,"xmin=%f xmax=%f zmin=%f zmax=%f",xmin,xmax,zmin,zmax);
	  k=fwrite_chd(fp,s,k);
	sprintf(s,"sx=%f sz=%f favg=%f ts=%f",sx,sz,favg,ts);
	  k=fwrite_chd(fp,s,k);
	sprintf(s,"wtype=%s stype=%s",wtype,stype);
	  k=fwrite_chd(fp,s,k);
	sprintf(s,"bc=%d,%d,%d,%d",bc[0],bc[1],bc[2],bc[3]);
	  k=fwrite_chd(fp,s,k);
	k=fwrite_chd(fp," ",k);
	k=fwrite_chd(fp,"Model file:",k);
	sprintf(s," %s",mfile);
	  k=fwrite_chd(fp,s,k);
	k=fwrite_chd(fp," ",k);
	if(qsw==1) k=fwrite_chd(fp,"Attenuation included",k);
	if(asw==1) k=fwrite_chd(fp,"Anisotropy included",k);
	while(k<=40) k=fwrite_chd(fp," ",k);
}	
	
int fwrite_chd(FILE *fp, char *s, int k)
{
	int i,n;	
	n=fprintf(fp,"C %s",s);
	for(i=1; i<80-n-1; ++i) fputs(" ",fp); 
	fprintf(fp,"%s\n"," ");
	return(k+1);
}
