/* Copyright (c) Colorado School of Mines, 2000.*/
/* All rights reserved.                       */

/* SUMAX: $Revision: 1.12 $ ; $Date: 1996/09/13 21:49:05 $			*/

#include "su.h"
#include "segy.h"

/*************************** self documentation **************************/
char *sdoc[] = 	{
"									",
" SUMAX - get trace by trace local/global maxima, minima, or absolute maximum",
"									",
" sumax <stdin >stdout [optional parameters] 				",
"									",
" Required parameters:							",
"	none								",
"									",
" Optional parameters: 							",
"	output=ascii 		write ascii data to outpar		",
"				=binary for binary floats to stdout	",
"				=segy for SEGY traces to stdout		",
"									",
"	mode=maxmin		output both minima and maxima		",
"				=max maxima only			",
"				=min minima only			",
"				=abs absolute maxima only      		",
"									",
"	verbose=0 		writes global quantities to outpar	",
"				=1 trace number, values, sample location",
"									",
"	outpar=/dev/tty		output parameter file; contains output	",
"					from verbose			",
"									",
" Examples: 								",
" For global max and min values:  sumax < segy_data			",
" For local and global max and min values:  sumax < segy_data verbose=1	",
" To plot values specified by mode:					",
"    sumax < segy_data output=binary mode=modeval | xgraph n=npairs	",
" To plot seismic data with the only values nonzero being those specified",
" by mode=modeval:							",
"    sumax < segy_data output=segy mode=modeval | suxwigb		",
"  									",
" Note:	while traces are counted from 1, sample values are counted from 0.",
"	Also, if multiple min, max, or abs max values exist on a trace,	",
"       only the first one is captured.					",
"									",
" See also: suxmax, supsmax						",
NULL};

/* Credits:
 *	CWP : John Stockwell (total rewrite)
 *	based on an original program by:
 *	SEP: Shuki Ronen
 *	CWP: Jack K. Cohen
 *
 * Trace header fields accessed: ns
 */
/**************** end self doc ***********************************/


segy tr;

int
main(int argc, char **argv)
{
	char *outpar;		/* name of file holding output parfile  */
	FILE *outparfp;		/* ... its file pointer			*/
	char *output;		/* format (ascii/binary/segy) of output	*/
	char *mode;		/* desired output maxmin/max/min/abs	*/
	int verbose;		/* flag to print extra information	*/

	int nt;			/* number of time points on trace	*/
	int ntr;		/* number of traces			*/

	int itr=0;		/* trace number	-- bumped at loop top 	*/
	float fitr=0;		/* trace number (as a float)		*/
	int it=0;		/* time point number			*/

	int gmaxit=0;		/* global max sample counter 		*/
	int gminit=0;		/* global min sample counter 		*/
	int gabsmaxit=0;	/* global abs max sample counter	*/

	int gmaxitr=1;		/* global max trace counter 		*/
	int gminitr=1;		/* global min trace counter 		*/
	int gabsmaxitr=1;	/* global abs max trace counter		*/

	int maxit=0;		/* sample number of max value		*/
	int minit=0;		/* sample number of min value		*/
	int absmaxit=0;		/* sample number of abs max value	*/

	float max;		/* max on a trace			*/
	float min;		/* min on a trace			*/
	float absmax;		/* absolute max on a trace		*/

	float gmax;		/* global max 				*/
	float gmin;		/* global min 				*/
	float gabsmax;		/* global absolute max	 		*/

	/* Initialize */
	initargs(argc, argv);
	requestdoc(1);

	/* Get optional parameters */
	if (!getparint("verbose", &verbose)) verbose=0;
	if (!getparstring("output", &output)) output="ascii";
	if (!(	STREQ(output, "ascii") || 
		STREQ(output, "binary") || 
		STREQ(output, "segy")
	)) err("%s unknown output selection", output); 

	if (!getparstring("mode", &mode)) mode="maxmin";
	if (!(	STREQ(mode, "maxmin")|| 
		STREQ(mode, "max") || 
		STREQ(mode, "min") || 
		STREQ(mode, "abs")
	)) err("%s unknown mode", mode); 

	if (STREQ(output, "binary") && STREQ(mode, "maxmin")) {
		mode="max";
		warn("outputting max values");
	}

	if (!getparstring("outpar", &outpar))  outpar = "/dev/tty" ;
	outparfp = efopen(outpar, "w");

	/* Get info from first trace */
	if (!gettr(&tr))  err("can't get first trace");
	nt = tr.ns;
	gmin = gmax = tr.data[0];
	gabsmax = ABS(tr.data[0]);

	/* Loop through data */
	do {
		itr++ ;

		/* find local/global max/min values and their locations */
		min = max = tr.data[0];
		absmax = ABS(tr.data[0]);
		tr.data[0] = 0.0; /* Zero first data value */
		for (it = 1; it < nt; ++it) {
			if (tr.data[it] > max) { /* Find max */
				max = tr.data[it];
				maxit = it;
			}

			if (tr.data[it] < min) { /* Find min */
				min = tr.data[it];
				minit = it;
			}

			if (ABS(tr.data[it]) > absmax) {/* Find absmax */
				absmax = ABS(tr.data[it]);
				absmaxit = it;
			}

			tr.data[it] = 0.0; /* zero data values */

		}
		/* prepare SEGY output */
		if (STREQ(output, "segy")) {
			if (STREQ(mode, "maxmin")) {
				tr.data[maxit] = max;
				tr.data[minit] = min;
			}
			if (STREQ(mode, "max")) {
				tr.data[maxit] = max;
			}
			if (STREQ(mode, "min")) {
				tr.data[minit] = min;
			}
			if (STREQ(mode, "abs")) {
				tr.data[absmaxit] = absmax;
			}
		}
		if (max > gmax) { /* Find global max */
			gmax = max;
			gmaxitr = itr;
			gmaxit  = maxit;
		}

		if (min < gmin) { /* Find global min */
			gmin = min;
			gminitr = itr;
			gminit  = minit;
		}
		if (absmax > gabsmax) { /* Find global min */
			gabsmax = absmax;
			gabsmaxitr = itr;
			gabsmaxit  = absmaxit;
		}
		/* Output ASCII */
		if (STREQ(output, "ascii") && verbose) { 
			/* max min */
			if (STREQ(mode, "maxmin")) {
				fprintf(outparfp, "%d %e at point %d\n",
						itr, max, maxit); 
				fprintf(outparfp, "   %e at point %d\n",
						min, minit); 
			}
			/* max */
			if (STREQ(mode, "max")) {
				fprintf(outparfp, "%d %e at point %d\n",
						itr, max, maxit); 
			}

			/* min */
			if (STREQ(mode, "min")) {
				fprintf(outparfp, "%d %e at point %d\n",
						itr, min, minit); 
			}

			/* abs max */
			if (STREQ(mode, "abs")) {
				fprintf(outparfp, "%d %e at point %d\n",
						itr, absmax, absmaxit); 
			}
		}

		/* Output raw BINARY pairs to stdout */ 
		fitr = itr; /* convert itr to floats, for binary output */
		if (STREQ(output, "binary")) {
			if (STREQ(mode, "max")){
				efwrite(&fitr, FSIZE, 1, stdout);
				efwrite(&max, FSIZE, 1, stdout);
			}
			if (STREQ(mode, "min")){
				efwrite(&fitr, FSIZE, 1, stdout);
				efwrite(&min, FSIZE, 1, stdout);
			}
			if (STREQ(mode, "abs")){
				efwrite(&fitr, FSIZE, 1, stdout);
				efwrite(&absmax, FSIZE, 1, stdout);
			}
		}
		/* Output SEGY data */
		if (STREQ(output, "segy"))  puttr (&tr);

	} while (gettr(&tr)); 

	ntr = itr ; /* Set ntr */

	/* Output global max min */
	if (STREQ(mode, "maxmin")) {
		if (verbose) { 
			fprintf(outparfp, 
			  "global max = %e, at trace = %d, at sample = %d\n",
				gmax, gmaxitr, gmaxit);
			fprintf(outparfp, 
			 "global min = %e, at trace = %d, at sample = %d\n", 
				gmin, gminitr, gminit);
		} else {
			fprintf(outparfp, "%e %e\n", gmax,gmin);
		}
	}

	/* Output global max */
	if (STREQ(mode, "max")) {
		if (verbose) { 
			fprintf(outparfp, 
			  "global max = %e, at trace = %d, at sample = %d\n",
				gmax, gmaxitr, gmaxit);
		} else {
			fprintf(outparfp, "%e\n", gmax);
		}
	}

	/* Output global min */
	if (STREQ(mode, "min")) {
		if (verbose) { 
			fprintf(outparfp, 
			  "global min = %e, at trace = %d, at sample = %d\n",
				gmin, gminitr, gminit);
		} else {
			fprintf(outparfp, "%e\n", gmin);
		}
	}

	/* Output global abs max */
	if (STREQ(mode, "abs")) {
		if (verbose) { 
			fprintf(outparfp, 
			 "global abs max = %e, at trace = %d, at sample = %d\n",
				gabsmax, gabsmaxitr, gabsmaxit);
		} else {
				fprintf(outparfp, "%e\n", gabsmax);
		}
	}

	/* If output=binary, print number of pairs to outpar */
	if (STREQ(output, "binary"))
		fprintf(outparfp, "npairs = %d\n", ntr);

	return EXIT_SUCCESS;
}
